/*
  appliance.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "appliance.h"

#include "dmp.h"
#include "sdt.h"
#include "rlp.h"
#include "rlp-udp.h"
#include "rlp-tcp.h"

namespace ACN {

Appliance::Appliance(UUID::uuid cid)
  : Component(cid)
{
  registerRlpVectorHandler(SDT::SDT_PROTOCOL_ID,
                           std::bind(&Appliance::rootSdtHandler, this,
                                     std::placeholders::_1));
  registerRlpVectorHandler(DMP::DMP_PROTOCOL_ID,
                           std::bind(&Appliance::rootDmpHandler, this,
                                     std::placeholders::_1));
};


/**
  EPI 17
*/
void Appliance::UdpStreamHandler(PDU::Stream stream) {
  // verify the UDP preamble
  RLP::UDP::preamble_t preamble(stream);

  // Implementations shall check the ACN Packet Identifier and preamble size.
  if (!preamble)
    stream->setstate(stream->rdstate() | std::ios_base::failbit);

  // implementations shall compute the size and position of the PDU block from
  // the preamble size and postamble size provided. ... ignoring any extra
  // octets in the preamble or postamble.
  for(int i = RLP::UDP::PREAMBLE_MINIMUM_SIZE; i < preamble.length; i++)
    stream->read8();

  if (!stream->good())
    return;
  auto block = PDU::readBlock<RLP::Pdu>(stream);
  if (stream->fail())
    return;

  for(auto const &root : *block) {
    RlpReceiver(root);
  }
}


/**
  EPI 33
*/
void Appliance::TcpStreamHandler(PDU::Stream stream) {
  // verify the TCP preamble
  RLP::TCP::preamble_t preamble(stream);

  // implementations shall check the ACN Packet Identifier. If the ACN Packet
  // Identifier is not correct the receiver shall close the connection.
  if (!preamble)
    stream->setstate(stream->rdstate() | std::ios_base::failbit);

  if (!stream->good())
    return;
  auto block = PDU::readBlock<RLP::Pdu>(stream);
  if (stream->fail())
    return;

  for(auto const &root : *block) {
    RlpReceiver(root);
  }
}


/**
  Dispatch a recieved RLP PDU to the appropriate vector handlers.
*/
void Appliance::RlpReceiver(std::shared_ptr<RLP::Pdu> root) {
  if (!rlp_vectors_.count(root->vector()))
    return;
  for(auto const &handler : rlp_vectors_[root->vector()]) {
    handler(root);
  }
}


/**
  Add callback handler for a given RLP vector.
*/
void Appliance::registerRlpVectorHandler(uint32_t vect,
                                         PDU::Handler<RLP::Pdu> handle) {
  rlp_vectors_[vect].push_back(handle);
}


/**
  Deregister RLP protocol handlers for the given vector.
*/
void Appliance::deregisterRlpVector(uint32_t vect) {
  rlp_vectors_.erase(vect);
}


/**
  Deregister _ALL_ RLP protocol handlers.
*/
void Appliance::deregisterRlpVector() {
  rlp_vectors_.clear();
}


/**

*/
void Appliance::rootSdtHandler(std::shared_ptr<RLP::Pdu> rlp) {

}


/**

*/
void Appliance::rootDmpHandler(std::shared_ptr<RLP::Pdu> rlp) {

}

}; // ACN
