/*
  appliance.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "component.h"
#include "rlp.h"
#include "sdt.h"
#include "pdu.h"
#include "../uuid/uuid.h"
#include <map>
#include <memory>
#include <vector>

// appliance: In DDL an appliance is a piece of equipment described by a root device and all its children and descendents. In DMP systems an appliance corresponds to a component that exposes one or more devices (since the rules require that all devices are descendants of a single root device). See also root device.

// component: The process, program or application corresponding to a single ACN endpoint. All messages in ACN are sent and received by a component. See [Arch] for a more complete definition.

// root device: An instance of a device described in DDL that has no parent device. See also appliance.


namespace ACN {

class Appliance
  : public Component
{
public:
  Appliance(UUID::uuid = UUID::uuid());

protected:
  // EPI 17
  virtual void UdpStreamHandler(PDU::Stream);
  // EPI 33
  virtual void TcpStreamHandler(PDU::Stream);

  void RlpReceiver(std::shared_ptr<RLP::Pdu>);
  void registerRlpVectorHandler(uint32_t, PDU::Handler<RLP::Pdu>);
  void deregisterRlpVector(uint32_t);
  void deregisterRlpVector();

  // process SDT frames
  virtual void rootSdtHandler(std::shared_ptr<RLP::Pdu>);
  // SDT 4.4 SDT Base Layer Messages
  // Join, Get Sessions, and Sessions are ad-hoc messages, that is, they are
  // not sent within channels. These messages identify their intended recipient
  // by CID.
  virtual void JoinSession() {};
  virtual void GetSessions() {};
  virtual void Sessions() {};

  // process DMP frames
  virtual void rootDmpHandler(std::shared_ptr<RLP::Pdu>);

private:
  std::map<uint32_t, std::vector<PDU::Handler<RLP::Pdu>>> rlp_vectors_;
  std::vector<std::shared_ptr<SDT::Session>> sessions_;
};

}; // ACN
