/*
  dmp.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "dmp.h"

namespace ACN {
namespace DMP {

address_type::address_type(uint8_t val) {
  z_reserved = (val >> 7) & 0b1;
  relative   = (val >> 6) & 0b1;
  type       = (data_type)((val >> 4) & 0b11);
  x_reserved = (val >> 2) & 0b11;
  width      = (address_length)(val & 0b11);
}

range::range(PDU::Stream stream, data_type t, address_length l) :
    incriment(0),
    count(0)
{
  address   = read(stream, l);
  if (t == SINGLE) return;
  incriment = read(stream, l);
  count     = read(stream, l);
}

uint32_t range::read(PDU::Stream stream, address_length length) {
  switch (length) {
    case ONE:
      return stream->read8();
    case TWO:
      return stream->read16();
    case FOUR:
      return stream->read32();
    default:
      return 0;
  }
}


Pdu::Pdu(PDU::Stream stream)
  : PDU::Pdu(stream, 1) // vectors are 1 octet
{
  if (stream->fail()) return;
  if (!buffer_->good()) return;

  if (flags_.hasHeader)
    setHeader(new address_type(buffer_->read8()));

  if (flags_.hasData) {
    switch(vector()) {
      case SET_PROPERTY:
        readSetData();
        break;
      default:
        break;
    }
  }
}

void Pdu::readSetData() {
  // header may be inherited. check that one exists
  if (!header()) {
    buffer_->setstate(buffer_->rdstate() | std::ios_base::failbit);
    return;
  }

  auto header = (address_type*)this->header();
  auto data = new dmp_set_data();

  while(buffer_->good()) {
    // Property Address
    range pr(buffer_, header->type, header->width);

    if (!buffer_->good() || pr.count * pr.incriment > buffer_->available()) {
      buffer_->setstate(buffer_->rdstate() | std::ios_base::failbit);
      break;
    }

    // Property Data
    std::vector<uint8_t> pd;
    pd.resize(pr.address, 0);
    for (uint32_t i = 0; i < pr.count * pr.incriment; i ++)
      pd.push_back(buffer_->read8());

    // Property Fields
    set_property fields(pr, pd);
    data->properties.push_back(fields);

    //set EOF if buffer insufficient for another property range
    if (buffer_->available() < header->width * (header->type == SINGLE? 1 : 3))
      buffer_->setstate(buffer_->rdstate() | std::ios_base::eofbit);
  }
  setData(data);
}

} // DMP
} // ACN
